////////////////////////////////////////////////////////////////////////////////
//                                                                            //
// IEController.cpp                                                           //
// Startprogramm fr berwachte Ausfhrung von Anwendungen.                   //
// (c) 2003 c't/Matthias Withopf                                              //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////


#define STRICT
#define WIN32_LEAN_AND_MEAN
#pragma warning(disable:4201)
#include <windows.h>
#pragma warning(default:4201)

#include <stdlib.h>
#include <strsafe.h>
#include "config.h"
#include "iecontroller.h"
#include "iecsupp.h"

static void       *AppInstance = NULL;
static const char *PrgmName    = APPNAME;

static char *ConfigBaseKeyName       = "Software\\c't\\IEController\\";
static char *PatchSupportDllFileName = "iecsupp.dll";
static char *PatchSupportFuncName    = "_Patch@20";

class TLauncher
  {
  public:
    typedef enum
      {
        lec_Ok,
        lec_OutOfMem,
        lec_InvalidArgOption,
        lec_CannotGetAppFileName,
        lec_InvalidArgument,
        lec_ProgramFileNotFound,
        lec_ProgramHasInvalidFormat,
        lec_CannotStartProgram,
        lec_ErrorLoadingSupportLibrary,
        lec_CannotGetSupportLibraryName,
        lec_InvalidSupportLibrary,
        lec_PatchCheckFailed,
        lec_CannotAllocateProcessPages,
        lec_CannotGetKernelDllHandle,
        lec_CannotGetKernelDllFuncs,
        lec_CannotWriteCodeFragment,
        lec_CannotWriteDataFragment,
        lec_CannotCreateRemoteThread,
        lec_ErrorWaitingForRemoteThread,
        lec_ErrorReadingThreadStatus,
        lec_RemoteThreadFailed,
        lec_NoAppEntry,
        lec_CannotPatchStartCode,
        lec_CannotGetEntryCode,
        lec_LAST
      } TLauncherErrorCode;
    TLauncherErrorCode LauncherErrorCode;
    TLauncher(char *CmdLine);
    ~TLauncher();
  private:
    bool OnlyLogging;
    bool ForceConfig;
    bool EnableLoggingAll;
    bool EnableLoggingDisabled;
    bool AskAlways;
    bool NoMouseCfg;
    bool ROConfig;
    bool GlobalConfig;
    bool SkipOpenFile;
    bool SkipWinSock;
    bool SkipNtDll;
    bool SkipOLE;
    bool SkipWinINet;
    char AppToRun[1024];
    char CfgName[80];
    bool  EvalArgLine(char *CmdLine,char **ArgsLine);
    PVOID GetEntryVA(char *FileName);
    bool  PatchProcess(char *AppFileName,PVOID AppEntry,HANDLE ProcessHandle,char *DllFileName);
    bool  StartProcess(char *FileName,char *ArgsLine);
  };
typedef TLauncher * PLauncher;

TLauncher::TLauncher(char *CmdLine)
{
  LauncherErrorCode     = lec_Ok;
  OnlyLogging           = false;
  ForceConfig           = false;
  EnableLoggingAll      = false;
  EnableLoggingDisabled = false;
  AskAlways             = false;
  NoMouseCfg            = false;
  ROConfig              = false;
  GlobalConfig          = false;
  SkipOpenFile          = false;
  SkipWinSock           = false;
  SkipNtDll             = false;
  SkipOLE               = false;
  SkipWinINet           = false;

  AppToRun[0] = '\0';
  CfgName[0]  = '\0';
  char *ArgsLine;
  if (!EvalArgLine(CmdLine,&ArgsLine))
    LauncherErrorCode = lec_InvalidArgOption;
  else
    {
      bool AppOk = true;
      if (!AppToRun[0])
        {
          if (!GetBrowserExecutableName(AppToRun,sizeof AppToRun))
            {
              LauncherErrorCode = lec_CannotGetAppFileName;
              AppOk = false;
            }
        }
      if (AppOk)
        {
          if (StartProcess(AppToRun,ArgsLine))
            LauncherErrorCode = lec_Ok;
          else
            {
              char tmp[1024],tmp1[80];
              TConfig::ResLoadString(AppInstance,strStartFailed,tmp1,sizeof tmp1);
              StringCbPrintf(tmp,sizeof tmp,tmp1,AppToRun);
              AlertMsg(tmp);
            }
        }
    }
}

TLauncher::~TLauncher()
{
}

bool TLauncher::EvalArgLine(char *CmdLine,char **ArgsLine)
{
  bool Result = true;
  char *p = CmdLine;
  while (*p)
    {
      while ((*p == ' ') || (*p == '\t'))
        ++p;
      if (*p != '-')
        break;
      char Arg[1024];
      char *p1 = Arg;
      bool WithinQuotes = false;
      while (*p)
        {
          char c = *p++;
          if (c == '"')
            WithinQuotes = !WithinQuotes;
          else
            {
              if ((c == ' ') || (c == '\t'))
                if (!WithinQuotes)
                  break;
              if ((p1 - Arg) < (sizeof Arg - 1))
                *p1++ = c;
            }
        }
      *p1 = '\0';

      if (!strcmp(Arg,"-OnlyLogging"))
        OnlyLogging = true;
      else if (!strcmp(Arg,"-Config"))
        ForceConfig = true;
      else if (!strcmp(Arg,"-EnableLogging"))
        EnableLoggingAll = true;
      else if (!strcmp(Arg,"-EnableLoggingDisabled"))
        EnableLoggingDisabled = true;
      else if (!strcmp(Arg,"-AskAlways"))
        AskAlways = true;
      else if (!strcmp(Arg,"-NoMouseCfg"))
        NoMouseCfg = true;
      else if (!strcmp(Arg,"-ROConfig"))
        ROConfig = true;
      else if (!strcmp(Arg,"-GlobalConfig"))
        GlobalConfig = true;
      else if (!strcmp(Arg,"-SkipOpenFile"))
        SkipOpenFile = true;
      else if (!strcmp(Arg,"-SkipWinSock"))
        SkipWinSock = true;
      else if (!strcmp(Arg,"-SkipNtDll"))
        SkipNtDll = true;
      else if (!strcmp(Arg,"-SkipOLE"))
        SkipOLE = true;
      else if (!strcmp(Arg,"-SkipWinINet"))
        SkipWinINet = true;
      else if (!strncmp(Arg,"-App:",5))
        StringCbCopy(AppToRun,sizeof AppToRun,Arg + 5);
      else if (!strncmp(Arg,"-Cfg:",5))
        StringCbCopy(CfgName,sizeof CfgName,Arg + 5);
      else
        {
          char tmp[1024],tmp1[80];
          TConfig::ResLoadString(AppInstance,strInvalidOption,tmp1,sizeof tmp1);
          StringCbPrintf(tmp,sizeof tmp,tmp1,Arg);
          AlertMsg(tmp);
          Result = false;
          break;
        }
    }
  if (ArgsLine)
    *ArgsLine = p;
  return Result;
}

bool TLauncher::PatchProcess(char *AppFileName,PVOID AppEntry,HANDLE ProcessHandle,char *DllFileName)
{
  bool Result = false;
  if (!AppEntry || !ProcessHandle || !DllFileName || !DllFileName[0])
    LauncherErrorCode = lec_InvalidArgument;
  else
    {
      char  FullDllFileName[1024];
      char *p;
      if (!GetFullPathName(DllFileName,sizeof FullDllFileName,FullDllFileName,&p))
        LauncherErrorCode = lec_CannotGetSupportLibraryName;
      else
        {
          SYSTEM_INFO SI;
          memset(&SI,0,sizeof SI);
          GetSystemInfo(&SI);
          DWORD PageSize = SI.dwPageSize;

          LPVOID Addr = NULL;
          LPVOID StartAddr = VirtualAllocEx(ProcessHandle,Addr,2 * PageSize,MEM_COMMIT,PAGE_EXECUTE_READWRITE);
          if (!StartAddr)
            LauncherErrorCode = lec_CannotAllocateProcessPages;
          else
            {
              PBYTE CodePage = (PBYTE)StartAddr;
              PBYTE DataPage = CodePage + PageSize;

              HINSTANCE Kernel32 = GetModuleHandle("kernel32.dll");
              if (!Kernel32)
                LauncherErrorCode = lec_CannotGetKernelDllHandle;
              else
                {
                  PVOID Addr_LoadLibraryA          = GetProcAddress(Kernel32,"LoadLibraryA");
                  PVOID Addr_GetProcAddress        = GetProcAddress(Kernel32,"GetProcAddress");
                  PVOID Addr_GetCurrentProcess     = GetProcAddress(Kernel32,"GetCurrentProcess");
                  PVOID Addr_WriteProcessMemory    = GetProcAddress(Kernel32,"WriteProcessMemory");
                  PVOID Addr_FlushInstructionCache = GetProcAddress(Kernel32,"FlushInstructionCache");
                  if (!Addr_LoadLibraryA || !Addr_GetProcAddress || !Addr_GetCurrentProcess || !Addr_WriteProcessMemory || !Addr_FlushInstructionCache)
                    LauncherErrorCode = lec_CannotGetKernelDllFuncs;
                  else
                    {
                      BYTE CodeFragment[1000];
                      BYTE DataFragment[4096];
                      memset(CodeFragment,0,sizeof CodeFragment);
                      memset(DataFragment,0,sizeof DataFragment);
                      PBYTE pc = CodeFragment;
                      PBYTE pd = DataFragment;

                      DWORD StatusOfs = ((DWORD)DataPage) + (pd - DataFragment);
                      *((DWORD *)pd) = 0; pd += sizeof(DWORD);

                      BYTE OrigEntryCode[8];
                      DWORD BytesRead;
                      if (!ReadProcessMemory(ProcessHandle,AppEntry,OrigEntryCode,sizeof OrigEntryCode,&BytesRead) || (BytesRead != sizeof OrigEntryCode))
                        LauncherErrorCode = lec_CannotGetEntryCode;
                      else
                        {
                          DWORD OrigEntryCodeOfs = ((DWORD)DataPage) + (pd - DataFragment);
                          memcpy(pd,OrigEntryCode,sizeof OrigEntryCode); pd += sizeof OrigEntryCode;

                          DWORD ScratchVarOfs = ((DWORD)DataPage) + (pd - DataFragment);
                          memset(pd,0,sizeof(DWORD)); pd += sizeof(DWORD);

                          uint l           = strlen(FullDllFileName) + 1;
                          DWORD DLLNameOfs = ((DWORD)DataPage) + (pd - DataFragment);
                          memcpy(pd,FullDllFileName,l); pd += l;

                          l                 = strlen(PatchSupportFuncName) + 1;
                          DWORD FuncNameOfs = ((DWORD)DataPage) + (pd - DataFragment);
                          memcpy(pd,PatchSupportFuncName,l); pd += l;

                          char ConfigKey[512];
                          StringCbCopy(ConfigKey,sizeof ConfigKey,ConfigBaseKeyName);
                          l = strlen(ConfigKey);
                          Str2KeyName(false,AppFileName,ConfigKey + l,sizeof ConfigKey - l);

                          l                  = strlen(ConfigKey) + 1;
                          DWORD ConfigKeyOfs = ((DWORD)DataPage) + (pd - DataFragment);
                          memcpy(pd,ConfigKey,l); pd += l;

                          l                = strlen(CfgName) + 1;
                          DWORD CfgNameOfs = ((DWORD)DataPage) + (pd - DataFragment);
                          memcpy(pd,CfgName,l); pd += l;

                          char AppFileName[1024];
                          GetModuleFileName(NULL,AppFileName,sizeof AppFileName);
                          l                    = strlen(AppFileName) + 1;
                          DWORD AppFileNameOfs = ((DWORD)DataPage) + (pd - DataFragment);
                          memcpy(pd,AppFileName,l); pd += l;

                          DWORD PatchFlags = pf_Default;
                          if (OnlyLogging)
                            PatchFlags |= pf_OnlyLogging;
                          if (ForceConfig)
                            PatchFlags |= pf_ForceConfig;
                          if (EnableLoggingAll)
                            PatchFlags |= pf_EnableLoggingAll;
                          if (EnableLoggingDisabled)
                            PatchFlags |= pf_EnableLoggingDisabled;
                          if (AskAlways)
                            PatchFlags |= pf_AskAlways;
                          if (ROConfig)
                            PatchFlags |= pf_ROConfig;
                          if (GlobalConfig)
                            PatchFlags |= pf_GlobalConfig;
                          if (SkipOpenFile)
                            PatchFlags |= pf_SkipOpenFile;
                          if (SkipWinSock)
                            PatchFlags |= pf_SkipWinSock;
                          if (SkipNtDll)
                            PatchFlags |= pf_SkipNtDll;
                          if (SkipOLE)
                            PatchFlags |= pf_SkipOLE;
                          if (SkipWinINet)
                            PatchFlags |= pf_SkipWinINet;

                          if (!NoMouseCfg)
                            if (GetAsyncKeyState(GetSystemMetrics(SM_SWAPBUTTON) ? VK_RBUTTON : VK_LBUTTON) < 0)
                              PatchFlags |= pf_ForceConfig;

                          *pc++ = 0x55;
                          *pc++ = 0x8B; *pc++ = 0xEC;
                          *pc++ = 0x56;
                          *pc++ = 0x57;
                          *pc++ = 0x68; *((DWORD *)pc) = DLLNameOfs; pc += sizeof(DWORD);
                          DWORD Ofs = (PBYTE)Addr_LoadLibraryA - (CodePage + (pc - CodeFragment)) - 5;
                          *pc++ = 0xE8; *((DWORD *)pc) = Ofs; pc += sizeof(DWORD);
                          *pc++ = 0x85; *pc++ = 0xC0;
                          *pc++ = 0x74; PBYTE L1 = pc; *pc++ = 0x00;
                          *pc++ = 0x8B; *pc++ = 0xF0;
                          *pc++ = 0x68; *((DWORD *)pc) = FuncNameOfs; pc += sizeof(DWORD);
                          *pc++ = 0x56;
                          Ofs = (PBYTE)Addr_GetProcAddress - (CodePage + (pc - CodeFragment)) - 5;
                          *pc++ = 0xE8; *((DWORD *)pc) = Ofs; pc += sizeof(DWORD);
                          *pc++ = 0x85; *pc++ = 0xC0;
                          *pc++ = 0x74; PBYTE L2 = pc; *pc++ = 0x00;
                          *pc++ = 0x8B; *pc++ = 0xF8;
                          *pc++ = 0x68; *((DWORD *)pc) = AppFileNameOfs; pc += sizeof(DWORD);
                          *pc++ = 0x68; *((DWORD *)pc) = CfgNameOfs; pc += sizeof(DWORD);
                          *pc++ = 0x68; *((DWORD *)pc) = ConfigKeyOfs; pc += sizeof(DWORD);
                          *pc++ = 0x68; *((DWORD *)pc) = PatchFlags; pc += sizeof(DWORD);
                          *pc++ = 0x56;
                          *pc++ = 0xFF; *pc++ = 0xD7;
                          *pc++ = 0x85; *pc++ = 0xC0;
                          *pc++ = 0x74; PBYTE L3 = pc; *pc++ = 0x00;
                          *pc++ = 0xC7; *pc++ = 0x05;
                          *((DWORD *)pc) = StatusOfs; pc += sizeof(DWORD);
                          *((DWORD *)pc) = 1; pc += sizeof(DWORD);
                          *L3 = (BYTE)(pc - L3 - 1);
                          *L2 = (BYTE)(pc - L2 - 1);
                          *L1 = (BYTE)(pc - L1 - 1);
                          *pc++ = 0x33; *pc++ = 0xC0;
                          *pc++ = 0x5F;
                          *pc++ = 0x5E;
                          *pc++ = 0x5D;
                          *pc++ = 0x60;
                          Ofs = (PBYTE)Addr_GetCurrentProcess - (CodePage + (pc - CodeFragment)) - 5;
                          *pc++ = 0xE8; *((DWORD *)pc) = Ofs; pc += sizeof(DWORD);
                          *pc++ = 0x8B; *pc++ = 0xF8;
                          *pc++ = 0x68; *((DWORD *)pc) = ScratchVarOfs; pc += sizeof(DWORD);
                          *pc++ = 0x68; *((DWORD *)pc) = sizeof OrigEntryCode; pc += sizeof(DWORD);
                          *pc++ = 0x68; *((DWORD *)pc) = OrigEntryCodeOfs; pc += sizeof(DWORD);
                          *pc++ = 0x68; *((DWORD *)pc) = (DWORD)AppEntry; pc += sizeof(DWORD);
                          *pc++ = 0x57;
                          Ofs = (PBYTE)Addr_WriteProcessMemory - (CodePage + (pc - CodeFragment)) - 5;
                          *pc++ = 0xE8; *((DWORD *)pc) = Ofs; pc += sizeof(DWORD);
                          *pc++ = 0x68; *((DWORD *)pc) = sizeof OrigEntryCode; pc += sizeof(DWORD);
                          *pc++ = 0x68; *((DWORD *)pc) = (DWORD)AppEntry; pc += sizeof(DWORD);
                          *pc++ = 0x57;
                          Ofs = (PBYTE)Addr_FlushInstructionCache - (CodePage + (pc - CodeFragment)) - 5;
                          *pc++ = 0xE8; *((DWORD *)pc) = Ofs; pc += sizeof(DWORD);
                          *pc++ = 0x61;
                          Ofs = (PBYTE)AppEntry - (CodePage + (pc - CodeFragment)) - 5;
                          *pc++ = 0xE9; *((DWORD *)pc) = Ofs; pc += sizeof(DWORD);

                          DWORD BytesWritten;
                          DWORD CodeSize = pc - CodeFragment;
                          if (!WriteProcessMemory(ProcessHandle,CodePage,CodeFragment,CodeSize,&BytesWritten) || (BytesWritten != CodeSize))
                            LauncherErrorCode = lec_CannotWriteCodeFragment;
                          else
                            {
                              DWORD DataSize = pd - DataFragment;
                              if (!WriteProcessMemory(ProcessHandle,DataPage,DataFragment,DataSize,&BytesWritten) || (BytesWritten != DataSize))
                                LauncherErrorCode = lec_CannotWriteDataFragment;
                              else
                                {
                                  FlushInstructionCache(ProcessHandle,CodePage,CodeSize);
                                  BYTE JmpOpCode[5];
                                  PBYTE pc = JmpOpCode;
                                  DWORD Ofs = (PBYTE)CodePage - ((PBYTE)AppEntry + (pc - JmpOpCode)) - 5;
                                  *pc++ = 0xE9; *((DWORD *)pc) = Ofs; pc += sizeof(DWORD);

                                  CodeSize = pc - JmpOpCode;
                                  if (!WriteProcessMemory(ProcessHandle,AppEntry,JmpOpCode,CodeSize,&BytesWritten) || (BytesWritten != CodeSize))
                                    LauncherErrorCode = lec_CannotPatchStartCode;
                                  else
                                    {
                                      FlushInstructionCache(ProcessHandle,AppEntry,CodeSize);
                                      Result = true;
                                    }
                                }
                           }
                       }
                    }
                }
            }
        }
    }
  return Result;
}

PVOID TLauncher::GetEntryVA(char *FileName)
{
  PVOID Result = NULL;
  if (FileName && FileName[0])
    {
      HANDLE F = CreateFile(FileName,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0);
      if (F != INVALID_HANDLE_VALUE)
        {
          BYTE  Buf[1024];
          DWORD ReadResult;
          if (ReadFile(F,Buf,2 * sizeof(BYTE),&ReadResult,NULL))
            {
              if ((Buf[0] == 'M') && (Buf[1] == 'Z'))
                {
                  LONG L = 0;
                  if (SetFilePointer(F,0x3C,&L,FILE_BEGIN))
                    {
                      DWORD PEHeaderOfs;
                      if (ReadFile(F,&PEHeaderOfs,sizeof PEHeaderOfs,&ReadResult,NULL))
                        {
                          L = 0;
                          if (SetFilePointer(F,PEHeaderOfs,&L,FILE_BEGIN))
                            {
                              if (ReadFile(F,Buf,24 * sizeof(BYTE),&ReadResult,NULL))
                                {
                                  if ((Buf[0] == 'P') && (Buf[1] == 'E') && (Buf[2] == '\0') && (Buf[3] == '\0'))
                                    {
                                      WORD Machine = *(PWORD)(Buf + 4);
                                      if (Machine == 0x014C)
                                        {
                                          WORD OptionalHeaderSize = *(PWORD)(Buf + 20);
                                          if (OptionalHeaderSize)
                                            {
                                              if (ReadFile(F,Buf,96 * sizeof(BYTE),&ReadResult,NULL))
                                                {
                                                  WORD MagicNumber = *(PWORD)(Buf + 0);
                                                  if (MagicNumber == 0x010B)
                                                    {
                                                      DWORD ImageBase = *(PDWORD)(Buf + 28);
                                                      DWORD EntryRVA  = *(PDWORD)(Buf + 16);
                                                      Result = (PVOID)(ImageBase + EntryRVA);
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
          CloseHandle(F);
        }
    }
  return Result;
}

bool TLauncher::StartProcess(char *FileName,char *ArgsLine)
{
  bool Result = false;
  if (!FileName || !FileName[0])
    LauncherErrorCode = lec_InvalidArgument;
  else
    {
      PVOID AppEntry = GetEntryVA(FileName);
      if (!AppEntry)
        LauncherErrorCode = lec_NoAppEntry;
      else
        {

          STARTUPINFO SI;
          memset(&SI,0,sizeof SI);
          SI.cb      = sizeof SI;
          SI.dwFlags = 0;
          PROCESS_INFORMATION PI;
          memset(&PI,0,sizeof PI);
          char ArgsLine2[16 * 1024];
          ArgsLine2[0] = '\0';
          if (ArgsLine && ArgsLine[0])
            {
              StringCbCopy(ArgsLine2,sizeof ArgsLine2,"\"");
              StringCbCat(ArgsLine2,sizeof ArgsLine2,FileName);
              StringCbCat(ArgsLine2,sizeof ArgsLine2,"\" ");
              StringCbCat(ArgsLine2,sizeof ArgsLine2,ArgsLine);
            }
          if (!CreateProcess(FileName,ArgsLine2[0] ? ArgsLine2 : NULL,NULL,NULL,FALSE,CREATE_SUSPENDED,NULL,NULL,&SI,&PI))
            {
              switch(GetLastError())
                {
                  case ERROR_FILE_NOT_FOUND:
                  case ERROR_PATH_NOT_FOUND:
                    LauncherErrorCode = lec_ProgramFileNotFound;
                    break;

                  case ERROR_BAD_FORMAT:
                    LauncherErrorCode = lec_ProgramHasInvalidFormat;
                    break;

                  default:
                    LauncherErrorCode = lec_CannotStartProgram;
                    break;
                }
            }
          else
            {
              char DllName[4 * 1024];
              DllName[0] = '\0';
              char AppName[4 * 1024];
              AppName[0] = '\0';
              if (GetModuleFileName(0,AppName,sizeof AppName))
                {
                  char *NamePart;
                  if (GetFullPathName(AppName,sizeof DllName,DllName,&NamePart))
                    {
                      if (NamePart)
                        {
                          *NamePart = '\0';
                          StringCbCat(DllName,sizeof DllName,PatchSupportDllFileName);
                        }
                      else
                        DllName[0] = '\0';
                    }
                }
              char *DLLFileName = DllName[0] ? DllName : PatchSupportDllFileName;

              HINSTANCE SuppDll = LoadLibrary(DLLFileName);
              if (!SuppDll)
                LauncherErrorCode = lec_ErrorLoadingSupportLibrary;
              else
                {
                  TPatchFunc PatchFunc = (TPatchFunc)GetProcAddress(SuppDll,PatchSupportFuncName);
                  if (!PatchFunc)
                    LauncherErrorCode = lec_InvalidSupportLibrary;
                  else
                    {
                      DWORD PatchFlags = pf_OnlyPatchCheck;
                      if (SkipOpenFile)
                        PatchFlags |= pf_SkipOpenFile;
                      if (SkipWinSock)
                        PatchFlags |= pf_SkipWinSock;
                      if (SkipNtDll)
                        PatchFlags |= pf_SkipNtDll;
                      if (SkipOLE)
                        PatchFlags |= pf_SkipOLE;
                      if (SkipWinINet)
                        PatchFlags |= pf_SkipWinINet;

                      if (!PatchFunc(SuppDll,PatchFlags,NULL,NULL,NULL))
                        LauncherErrorCode = lec_PatchCheckFailed;
                      else
                        {
                          if (PatchProcess(FileName,AppEntry,PI.hProcess,DLLFileName))
                            {
                              Result = true;
                            }
                        }
                    }
                  FreeLibrary(SuppDll);
                }
              ResumeThread(PI.hThread);
              CloseHandle(PI.hThread);

              CloseHandle(PI.hProcess);
            }
        }
    }
  return Result;
}

int CALLBACK WinMain(HINSTANCE hInstance,HINSTANCE /*hPrevInstance*/,LPSTR lpszCmdLine,int /*nCmdShow*/)
{
  int Result = 0;
  AppInstance = hInstance;
  PLauncher PL = new TLauncher(lpszCmdLine);
  if (!PL)
    Result = TLauncher::lec_OutOfMem;
  else
    {
      if (PL->LauncherErrorCode)
        Result = PL->LauncherErrorCode;
      delete PL;
    }
  if (Result)
    {
      UINT MsgId = 0;
      switch(Result)
        {
          case TLauncher::lec_OutOfMem:                    MsgId = strOutOfMem;                    break;
          case TLauncher::lec_InvalidArgOption:            MsgId = strInvalidArgOption;            break;
          case TLauncher::lec_CannotGetAppFileName:        MsgId = strCannotGetAppFileName;        break;
          case TLauncher::lec_InvalidArgument:             MsgId = strInvalidArgument;             break;
          case TLauncher::lec_ProgramFileNotFound:         MsgId = strProgramFileNotFound;         break;
          case TLauncher::lec_ProgramHasInvalidFormat:     MsgId = strProgramHasInvalidFormat;     break;
          case TLauncher::lec_CannotStartProgram:          MsgId = strCannotStartProgram;          break;
          case TLauncher::lec_ErrorLoadingSupportLibrary:  MsgId = strErrorLoadingSupportLibrary;  break;
          case TLauncher::lec_CannotGetSupportLibraryName: MsgId = strCannotGetSupportLibraryName; break;
          case TLauncher::lec_InvalidSupportLibrary:       MsgId = strInvalidSupportLibrary;       break;
          case TLauncher::lec_PatchCheckFailed:            MsgId = strPatchCheckFailed;            break;
          case TLauncher::lec_CannotAllocateProcessPages:  MsgId = strCannotAllocateProcessPages;  break;
          case TLauncher::lec_CannotGetKernelDllHandle:    MsgId = strCannotGetKernelDllHandle;    break;
          case TLauncher::lec_CannotGetKernelDllFuncs:     MsgId = strCannotGetKernelDllFuncs;     break;
          case TLauncher::lec_CannotWriteCodeFragment:     MsgId = strCannotWriteCodeFragment;     break;
          case TLauncher::lec_CannotWriteDataFragment:     MsgId = strCannotWriteDataFragment;     break;
          case TLauncher::lec_CannotCreateRemoteThread:    MsgId = strCannotCreateRemoteThread;    break;
          case TLauncher::lec_ErrorWaitingForRemoteThread: MsgId = strErrorWaitingForRemoteThread; break;
          case TLauncher::lec_ErrorReadingThreadStatus:    MsgId = strErrorReadingThreadStatus;    break;
          case TLauncher::lec_RemoteThreadFailed:          MsgId = strRemoteThreadFailed;          break;
          case TLauncher::lec_NoAppEntry:                  MsgId = strNoAppEntry;                  break;
          case TLauncher::lec_CannotPatchStartCode:        MsgId = strCannotPatchStartCode;        break;
          case TLauncher::lec_CannotGetEntryCode:          MsgId = strCannotGetEntryCode;          break;
        }
      char tmp[256];
      TConfig::ResLoadString(AppInstance,MsgId,tmp,sizeof tmp);
      AlertMsg(tmp);
    }
  return Result;
}

